from multiprocessing.sharedctypes import Value
import numpy as np
import matplotlib.pyplot as plt
from pyro import param
from scipy.stats.qmc import Sobol, scale
import math
import torch



################################################################################################
######################################## UQ Simple beam example ################################
def  uq_simple_beam_F(n=100, X = None, 
    fidelity = 0,noise_std = 0.0, random_state = None, shuffle = True):

    if random_state is not None:
        np.random.seed(random_state)

    dx = 5
    l_bound = [.1,.2,4,5e+9,1e+4]
    u_bound = [.2,.4,6,35e+9,1.4e+4]

    out_flag = 0
    if X is None:
        sobolset = Sobol(d=dx, seed = random_state)
        X = sobolset.random(2 ** (np.log2(n) + 1).astype(int))[:n, :]
        X = scale(X, l_bounds=l_bound, u_bounds=u_bound)
        out_flag = 1
    if type(X) != np.ndarray:
        X = np.array(X)
    

    if fidelity == 0:
        X[..., 0]=0.15
        X[..., 1]=0.3
        X[..., 2]=5
        X[..., 3]=30e+9
        X[..., 4]=1.2e+04

        y = (5/32)*(X[..., 4]*X[..., 2]**4)/(X[..., 3]*X[..., 0]*X[..., 1]**3)
        
    elif fidelity == 1:
        y = (5/32)*(X[..., 4]*X[..., 2]**4)/(X[..., 3]*X[..., 0]*X[..., 1]**3)

    else:
        raise ValueError('only 3 fidelities of 0,1,2 have been implemented ')

    if X is None:
        if shuffle:
            index = np.random.randint(0, len(y), size = len(y))
            X = X[index,...]
            y = y[index]

    if noise_std > 0.0:
        if out_flag == 1:
            return X, y + np.random.randn(*y.shape) * noise_std
        else:
            return y       
    else:
        if out_flag == 1:
            return X, y
        else:
            return y


################################# Multi-fidelity-1D_Inverse for Calibration ####################################
def  uq_simple_beam(X = None,
    n={'0': 50, '1': 100},
    noise_std={'0': 0.0, '1': 0.0},
    random_state = None, shuffle = True):
    if X is None:
        X_list = []
        y_list = []
        for level, num in n.items():
            if level  in ['0','1','2'] and num > 0:
                X, y =  uq_simple_beam_F(n=num, fidelity = int(level), 
                    noise_std= noise_std[level], random_state = random_state)
                X = np.hstack([X, np.ones(num).reshape(-1, 1) * float(level)])
                X_list.append(X)
                y_list.append(y)
            else:
                raise ValueError('Wrong label, should be h, l1, l2 or l3')
        return np.vstack([*X_list]), np.hstack(y_list)
    else:
        fidelity = [i for i in n.keys()]
        y_list = []
        if type(X) == np.ndarray:
            X = torch.tensor(X)
        for f in fidelity:
            # Find the index for each fidelity then call the wing
            index = [i[0] for i in torch.argwhere(X[...,-1]== int(f))]
            y_list.append(uq_simple_beam_F(X=X[index, 0:-1], fidelity=int(f),  
                noise_std= noise_std[f]))
        return torch.tensor(np.hstack(y_list))
    
################################## Calibration_1D Poly #########################################
def calibration_1D_Inverse(n=100, X = None, 
    fidelity = 0,noise_std = 0.0, random_state = None, shuffle = True):

    if random_state is not None:
        np.random.seed(random_state)

    dx = 2
    l_bound = [-.5, -1]
    u_bound = [.5, 2]
    out_flag = 0
    if X is None:
        sobolset = Sobol(d=dx, seed = random_state)
        X = sobolset.random(2 ** (np.log2(n) + 1).astype(int))[:n, :]
        X = scale(X, l_bounds=l_bound, u_bounds=u_bound)
        out_flag = 1
    if type(X) != np.ndarray:
        X = np.array(X)

    # This is the output
    if fidelity == 0:
        X[..., 0]=.1
        y = 1/(X[..., 0] * X[..., 1]**3 + X[..., 1]**2 + X[..., 1] +1)
    elif fidelity == 1:
        y = 1/(X[..., 0] * X[..., 1]**3 + X[..., 1]**2 + X[..., 1] +1)
    elif fidelity == 2:
        y =1/(X[..., 0] * X[..., 1]**2 + X[..., 1] +1)

    else:
        raise ValueError('only 3 fidelities of 0,1,2 have been implemented ')

    if X is None:
        if shuffle:
            index = np.random.randint(0, len(y), size = len(y))
            X = X[index,...]
            y = y[index]


    if noise_std > 0.0:
        if out_flag == 1:
            return X, y + np.random.randn(*y.shape) * noise_std
        else:
            return y       
    else:
        if out_flag == 1:
            return X, y
        else:
            return y


################################# Multi-fidelity-1D_Inverse for Calibration ####################################
def Calibration_1D_poly_invers(X = None,
    n={'0': 50, '1': 100, '2': 100},
    noise_std={'0': 0.0, '1': 0.0, '2': 0.0},
    random_state = None, shuffle = True):
    if X is None:
        X_list = []
        y_list = []
        for level, num in n.items():
            if level  in ['0','1','2'] and num > 0:
                X, y = calibration_1D_Inverse(n=num, fidelity = int(level), 
                    noise_std= noise_std[level], random_state = random_state)
                X = np.hstack([X, np.ones(num).reshape(-1, 1) * float(level)])
                X_list.append(X)
                y_list.append(y)
            else:
                raise ValueError('Wrong label, should be h, l1, l2 or l3')
        return np.vstack([*X_list]), np.hstack(y_list)
    else:
        fidelity = [i for i in n.keys()]
        y_list = []
        if type(X) == np.ndarray:
            X = torch.tensor(X)
        for f in fidelity:
            # Find the index for each fidelity then call the wing
            index = [i[0] for i in torch.argwhere(X[...,-1]== int(f))]
            y_list.append(calibration_1D_Inverse(X=X[index, 0:-1], fidelity=int(f),  
                noise_std= noise_std[f]))
        return torch.tensor(np.hstack(y_list))
    
###############################################################################################
################################## Calibration_1D Poly #########################################
def calibration_1D(n=100, X = None, 
    fidelity = 0,noise_std = 0.0, random_state = None, shuffle = True):

    if random_state is not None:
        np.random.seed(random_state)

    dx = 2
    l_bound = [-2, -2]
    u_bound = [2, 3]
    out_flag = 0
    if X is None:
        sobolset = Sobol(d=dx, seed = random_state)
        X = sobolset.random(2 ** (np.log2(n) + 1).astype(int))[:n, :]
        X = scale(X, l_bounds=l_bound, u_bounds=u_bound)
        out_flag = 1
    if type(X) != np.ndarray:
        X = np.array(X)

    if fidelity == 0:
        X[..., 0]=.1
        y = X[..., 0] * X[..., 1]**3 + X[..., 1]**2 + X[..., 1] +1
    elif fidelity == 1:
    # This is the output
        y =X[..., 0] * X[..., 1]**3 + X[..., 1]**2 + X[..., 1] +1

    elif fidelity == 2:
    # This is level 2 in Tammers paper
        y =X[..., 0] * X[..., 1]**3 + X[..., 1]**2 +1

    else:
        raise ValueError('only 3 fidelities of 0,1,2 have been implemented ')

    if X is None:
        if shuffle:
            index = np.random.randint(0, len(y), size = len(y))
            X = X[index,...]
            y = y[index]


    if noise_std > 0.0:
        if out_flag == 1:
            return X, y + np.random.randn(*y.shape) * noise_std
        else:
            return y       
    else:
        if out_flag == 1:
            return X, y
        else:
            return y


################################# Multi-fidelity-1D for Calibration ####################################
def calibration_1D_poly(X = None,
    n={'0': 50, '1': 100, '2': 100, '3': 100},
    noise_std={'0': 0.0, '1': 0.0, '2': 0.0, '3': 0.0},
    random_state = None, shuffle = True):
    if X is None:
        X_list = []
        y_list = []
        for level, num in n.items():
            if level  in ['0','1','2'] and num > 0:
                X, y = calibration_1D(n=num, fidelity = int(level), 
                    noise_std= noise_std[level], random_state = random_state)
                X = np.hstack([X, np.ones(num).reshape(-1, 1) * float(level)])
                X_list.append(X)
                y_list.append(y)
            else:
                raise ValueError('Wrong label, should be h, l1, l2 or l3')
        return np.vstack([*X_list]), np.hstack(y_list)
    else:
        fidelity = [i for i in n.keys()]
        y_list = []
        if type(X) == np.ndarray:
            X = torch.tensor(X)
        for f in fidelity:
            # Find the index for each fidelity then call the wing
            index = [i[0] for i in torch.argwhere(X[...,-1]== int(f))]
            y_list.append(calibration_1D(X=X[index, 0:-1], fidelity=int(f),  
                noise_std= noise_std[f]))
        return torch.tensor(np.hstack(y_list))
    
################################## Calibration_1D Poly #########################################
def calibration_3_sources_sin(n=100, X = None, 
    fidelity = 0,noise_std = 0.0, random_state = None, shuffle = True):

    if random_state is not None:
        np.random.seed(random_state)

    dx = 2
    l_bound = [0, -1]
    u_bound = [7, 1]
    out_flag = 0
    if X is None:
        sobolset = Sobol(d=dx, seed = random_state)
        X = sobolset.random(2 ** (np.log2(n) + 1).astype(int))[:n, :]
        X = scale(X, l_bounds=l_bound, u_bounds=u_bound)
        out_flag = 1
    if type(X) != np.ndarray:
        X = np.array(X)

    if fidelity == 0:
        X[..., 0]=1
        y = np.sin(X[..., 0]*X[..., 1]) + np.sin(2*X[..., 0]*X[..., 1])
    elif fidelity == 1:
    # This is the output
        y =np.sin(X[..., 0]*X[..., 1])

    elif fidelity == 2:
    # This is level 2 in Tammers paper
        y =np.sin(X[..., 0]*X[..., 1]) + np.sin(2*X[..., 0]*X[..., 1])

    else:
        raise ValueError('only 3 fidelities of 0,1,2 have been implemented ')

    if X is None:
        if shuffle:
            index = np.random.randint(0, len(y), size = len(y))
            X = X[index,...]
            y = y[index]


    if noise_std > 0.0:
        if out_flag == 1:
            return X, y + np.random.randn(*y.shape) * noise_std
        else:
            return y       
    else:
        if out_flag == 1:
            return X, y
        else:
            return y


################################# Multi-fidelity-1D for Calibration ####################################
def calibration_1D_3_sources_sin(X = None,
    n={'0': 50, '1': 100, '2': 100, '3': 100},
    noise_std={'0': 0.0, '1': 0.0, '2': 0.0, '3': 0.0},
    random_state = None, shuffle = True):
    if X is None:
        X_list = []
        y_list = []
        for level, num in n.items():
            if level  in ['0','1','2'] and num > 0:
                X, y = calibration_3_sources_sin(n=num, fidelity = int(level), 
                    noise_std= noise_std[level], random_state = random_state)
                X = np.hstack([X, np.ones(num).reshape(-1, 1) * float(level)])
                X_list.append(X)
                y_list.append(y)
            else:
                raise ValueError('Wrong label, should be h, l1, l2 or l3')
        return np.vstack([*X_list]), np.hstack(y_list)
    else:
        fidelity = [i for i in n.keys()]
        y_list = []
        if type(X) == np.ndarray:
            X = torch.tensor(X)
        for f in fidelity:
            # Find the index for each fidelity then call the wing
            index = [i[0] for i in torch.argwhere(X[...,-1]== int(f))]
            y_list.append(calibration_3_sources_sin(X=X[index, 0:-1], fidelity=int(f),  
                noise_std= noise_std[f]))
        return torch.tensor(np.hstack(y_list))
    
################################## Calibration_1D_poly_2_Sources #########################################
def calibration_1D_2_Sources(n=100, X = None, 
    fidelity = 0,noise_std = 0.0, random_state = None, shuffle = True):

    if random_state is not None:
        np.random.seed(random_state)

    dx = 2
    l_bound = [-2, -2]
    u_bound = [2, 3]
    out_flag = 0
    if X is None:
        sobolset = Sobol(d=dx, seed = random_state)
        X = sobolset.random(2 ** (np.log2(n) + 1).astype(int))[:n, :]
        X = scale(X, l_bounds=l_bound, u_bounds=u_bound)
        out_flag = 1
    if type(X) != np.ndarray:
        X = np.array(X)
    # Wp = X[..., 9]
    # This is the output

    if fidelity == 0:
        mean = 0  # Mean of the Gaussian distribution
        stdd= .1  # Variance of the Gaussian distribution
        X[..., 0] = np.random.normal(mean, stdd,X[..., 0].shape)

        y = X[..., 0] * X[..., 1]**3 +X[..., 1]**2 +1
    elif fidelity == 1:
    # This is the output
        y =X[..., 0] * X[..., 1]**3 +X[..., 1]**2 +1
    else:
        raise ValueError('only 3 fidelities of 0,1,2 have been implemented ')

    if X is None:
        if shuffle:
            index = np.random.randint(0, len(y), size = len(y))
            X = X[index,...]
            y = y[index]

    if noise_std > 0.0:
        if out_flag == 1:
            return X, y + np.random.randn(*y.shape) * noise_std
        else:
            return y       
    else:
        if out_flag == 1:
            return X, y
        else:
            return y


################################# Multi-fidelity-1D for Calibration ####################################
def calibration_1D_poly_2_Sources(X = None,
    n={'0': 50, '1': 100},
    noise_std={'0': 0.0, '1': 0.0},
    random_state = None, shuffle = True):
    if X is None:
        X_list = []
        y_list = []
        for level, num in n.items():
            if level  in ['0','1'] and num > 0:
                X, y = calibration_1D_2_Sources(n=num, fidelity = int(level), 
                    noise_std= noise_std[level], random_state = random_state)
                X = np.hstack([X, np.ones(num).reshape(-1, 1) * float(level)])
                X_list.append(X)
                y_list.append(y)
            else:
                raise ValueError('Wrong label, should be h, l1, l2 or l3')
        return np.vstack([*X_list]), np.hstack(y_list)
    else:
        fidelity = [i for i in n.keys()]
        y_list = []
        if type(X) == np.ndarray:
            X = torch.tensor(X)
        for f in fidelity:
            # Find the index for each fidelity then call the wing
            index = [i[0] for i in torch.argwhere(X[...,-1]== int(f))]
            y_list.append(calibration_1D_2_Sources(X=X[index, 0:-1], fidelity=int(f),  
                noise_std= noise_std[f]))
        return torch.tensor(np.hstack(y_list))

################################# Multi-fidelity-1D for Calibration ####################################
def calibration_1D_cos(X = None,
    n={'0': 50, '1': 100},
    noise_std={'0': 0.0, '1': 0.0},
    random_state = None, shuffle = True):
    if X is None:
        X_list = []
        y_list = []
        for level, num in n.items():
            if level  in ['0','1'] and num > 0:
                X, y =calibration_f_cos(n=num, fidelity = int(level), 
                    noise_std= noise_std[level], random_state = random_state)
                X = np.hstack([X, np.ones(num).reshape(-1, 1) * float(level)])
                X_list.append(X)
                y_list.append(y)
            else:
                raise ValueError('Wrong label, should be h, l1')
        return np.vstack([*X_list]), np.hstack(y_list)
    else:
        fidelity = [i for i in n.keys()]
        y_list = []
        if type(X) == np.ndarray:
            X = torch.tensor(X)
        for f in fidelity:
            # Find the index for each fidelity then call the wing
            index = [i[0] for i in torch.argwhere(X[...,-1]== int(f))]
            y_list.append(calibration_f_cos(X=X[index, 0:-1], fidelity=int(f),  
                noise_std= noise_std[f]))
        return torch.tensor(np.hstack(y_list))
    
################################## Calibration Cos() #########################################
def calibration_f_cos(n=100, X = None, 
    fidelity = 0,noise_std = 0.0, random_state = None, shuffle = True):
    if random_state is not None:
        np.random.seed(random_state)

    dx = 3
    l_bound =[-2,0,-1]
    u_bound = [2,10,1]
    out_flag = 0
    if X is None:
        sobolset = Sobol(d=dx, seed = random_state)
        X = sobolset.random(2 ** (np.log2(n) + 1).astype(int))[:n, :]
        X = scale(X, l_bounds=l_bound, u_bounds=u_bound)
        out_flag = 1
    if type(X) != np.ndarray:
        X = np.array(X)
    if fidelity == 0:
        X[..., 0]=1
        X[..., 1]=.5
        
        y = 2*X[..., 0]+np.cos(X[..., 1]*X[..., 2]) + 1*np.cos(10*X[..., 1]*X[..., 2])

    elif fidelity == 1:

        y =  2*X[..., 0] +np.cos(X[..., 1]*X[..., 2])

    else:
        raise ValueError('only 2 fidelities of 0,1 have been implemented ')

    if X is None:
        if shuffle:
            index = np.random.randint(0, len(y), size = len(y))
            X = X[index,...]
            y = y[index]


    if noise_std > 0.0:
        if out_flag == 1:
            return X, y + np.random.randn(*y.shape) * noise_std
        else:
            return y       
    else:
        if out_flag == 1:
            return X, y
        else:
            return y

################################## Calibration SIN() #########################################
def calibration_sin_M(n=100, X = None, 
    fidelity = 0,noise_std = 0.0, random_state = None, shuffle = True):

    if random_state is not None:
        np.random.seed(random_state)

    dx = 2
    l_bound =[.8, -2.5]
    u_bound = [1.6, 2.5]
    out_flag = 0
    if X is None:
        sobolset = Sobol(d=dx, seed = random_state)
        X = sobolset.random(2 ** (np.log2(n) + 1).astype(int))[:n, :]
        X = scale(X, l_bounds=l_bound, u_bounds=u_bound)
        out_flag = 1
    if type(X) != np.ndarray:
        X = np.array(X)
    if fidelity == 0:
        X[..., 0]=1.2
        y = np.sin(X[..., 0] * X[..., 1]) + 0.1 * X[..., 1] 

    elif fidelity == 1:
    # This is the output
        y =np.sin(X[..., 0] * X[..., 1]) 

    else:
        raise ValueError('only 2 fidelities of 0,1 have been implemented ')

    if X is None:
        if shuffle:
            index = np.random.randint(0, len(y), size = len(y))
            X = X[index,...]
            y = y[index]

    if noise_std > 0.0:
        if out_flag == 1:
            return X, y + np.random.randn(*y.shape) * noise_std
        else:
            return y       
    else:
        if out_flag == 1:
            return X, y
        else:
            return y


################################# Multi-fidelity-1D for 2 Calibration Parameters####################################
def calibration_1D_sin_koh_example(X = None,
    n={'0': 50, '1': 100},
    noise_std={'0': 0.0, '1': 0.0},
    random_state = None, shuffle = True):
    if X is None:
        X_list = []
        y_list = []
        for level, num in n.items():
            if level  in ['0','1'] and num > 0:
                X, y =calibration_sin_M(n=num, fidelity = int(level), 
                    noise_std= noise_std[level], random_state = random_state)
                X = np.hstack([X, np.ones(num).reshape(-1, 1) * float(level)])
                X_list.append(X)
                y_list.append(y)
            else:
                raise ValueError('Wrong label, should be h, l1')
        return np.vstack([*X_list]), np.hstack(y_list)
    else:
        fidelity = [i for i in n.keys()]
        y_list = []
        if type(X) == np.ndarray:
            X = torch.tensor(X)
        for f in fidelity:
            # Find the index for each fidelity then call the wing
            index = [i[0] for i in torch.argwhere(X[...,-1]== int(f))]
            y_list.append(calibration_sin_M(X=X[index, 0:-1], fidelity=int(f),  
                noise_std= noise_std[f]))
        return torch.tensor(np.hstack(y_list))
    
################################## Calibration SIN() #########################################
def calibration_f_sin(n=100, X = None, 
    fidelity = 0,noise_std = 0.0, random_state = None, shuffle = True):


    if random_state is not None:
        np.random.seed(random_state)

    dx = 2
    l_bound =[-5, 0]
    u_bound = [15, 1]
    out_flag = 0
    if X is None:
        sobolset = Sobol(d=dx, seed = random_state)
        X = sobolset.random(2 ** (np.log2(n) + 1).astype(int))[:n, :]
        X = scale(X, l_bounds=l_bound, u_bounds=u_bound)
        out_flag = 1
    if type(X) != np.ndarray:
        X = np.array(X)
    if fidelity == 0:
        X[..., 0]=2
        y = np.sin(X[..., 0]*X[..., 1]) + np.sin(5*X[..., 0]*X[..., 1])

    elif fidelity == 1:
    # This is the output
        y =np.sin(X[..., 0]*X[..., 1])

    else:
        raise ValueError('only 2 fidelities of 0,1 have been implemented ')

    if X is None:
        if shuffle:
            index = np.random.randint(0, len(y), size = len(y))
            X = X[index,...]
            y = y[index]


    if noise_std > 0.0:
        if out_flag == 1:
            return X, y + np.random.randn(*y.shape) * noise_std
        else:
            return y       
    else:
        if out_flag == 1:
            return X, y
        else:
            return y


################################# Multi-fidelity-1D for 2 Calibration Parameters####################################
def calibration_1D_sin(X = None,
    n={'0': 50, '1': 100},
    noise_std={'0': 0.0, '1': 0.0},
    random_state = None, shuffle = True):
    if X is None:
        X_list = []
        y_list = []
        for level, num in n.items():
            if level  in ['0','1'] and num > 0:
                X, y =calibration_f_sin(n=num, fidelity = int(level), 
                    noise_std= noise_std[level], random_state = random_state)
                X = np.hstack([X, np.ones(num).reshape(-1, 1) * float(level)])
                X_list.append(X)
                y_list.append(y)
            else:
                raise ValueError('Wrong label, should be h, l1')
        return np.vstack([*X_list]), np.hstack(y_list)
    else:
        fidelity = [i for i in n.keys()]
        y_list = []
        if type(X) == np.ndarray:
            X = torch.tensor(X)
        for f in fidelity:
            # Find the index for each fidelity then call the wing
            index = [i[0] for i in torch.argwhere(X[...,-1]== int(f))]
            y_list.append(calibration_f_sin(X=X[index, 0:-1], fidelity=int(f),  
                noise_std= noise_std[f]))
        return torch.tensor(np.hstack(y_list))
    
################################## Calibration Wing #########################################
################################## Wing #########################################
def calibration_f_wing(n=100, X = None, 
    fidelity = 0,noise_std = 0.0, random_state = None, shuffle = True):


    if random_state is not None:
        np.random.seed(random_state)

    dx = 10
    l_bound = [150, 220, 6, -10, 16, 0.7, 0.08, 2, 1700, 0.025]
    u_bound = [200, 300, 10, 10, 45, 0.9, 0.20, 3.5, 2500, 0.08]
    out_flag = 0
    if X is None:
        sobolset = Sobol(d=dx, seed = random_state)
        X = sobolset.random(2 ** (np.log2(n) + 1).astype(int))[:n, :]
        X = scale(X, l_bounds=l_bound, u_bounds=u_bound)
        out_flag = 1
    if type(X) != np.ndarray:
        X = np.array(X)
    Sw = X[..., 0]
    Wfw = X[..., 1]
    A = X[..., 2]
    Gama = X[..., 3] * (np.pi/180.0)
    q = X[..., 4]
    lamb = X[..., 5]
    tc = X[..., 6]
    Nz = X[..., 7]
    Wdg = X[..., 8]
    Wp = X[..., 9]
    # This is the output

    if fidelity == 0:
        X[..., 4]=40
        X[..., 5]=0.85
        X[..., 6]=0.17
        X[..., 7]=3
        y = 0.036 * Sw**0.758 * Wfw**0.0035 * (A/(np.cos(Gama)) ** 2) ** 0.6 * \
            q**0.006*lamb**0.04 * ((100 * tc)/(np.cos(Gama)))**(-0.3) *\
            (Nz * Wdg) ** 0.49 + Sw * Wp
    elif fidelity == 1:
    # This is the output
        y = 0.036 * Sw**0.758 * Wfw**0.0035 * (A/(np.cos(Gama)) ** 2) ** 0.6 * \
            q**0.006*lamb**0.04 * ((100 * tc)/(np.cos(Gama)))**(-0.3)\
            * (Nz * Wdg) ** 0.49 + 1 * Wp 

    elif fidelity == 2:
    # This is level 2 in Tammers paper
        y = 0.036 * Sw**0.8 * Wfw**0.0035 * (A/(np.cos(Gama)) ** 2) ** 0.6 * \
            q**0.006*lamb**0.04 * ((100 * tc)/(np.cos(Gama)))**(-0.3)\
            * (Nz * Wdg) ** 0.49 + 1 * Wp

    elif fidelity == 3:
    # This is level 2 in Tammers paper
        y = 0.036 * Sw**0.9 * Wfw**0.0035 * (A/(np.cos(Gama)) ** 2) ** 0.6 * \
            q**0.006*lamb**0.04 * ((100 * tc)/(np.cos(Gama)))**(-0.3)\
            * (Nz * Wdg) ** 0.49 + 0 * Wp
    else:
        raise ValueError('only 4 fidelities of 0,1,2,3 have been implemented ')

    if X is None:
        if shuffle:
            index = np.random.randint(0, len(y), size = len(y))
            X = X[index,...]
            y = y[index]


    if noise_std > 0.0:
        if out_flag == 1:
            return X, y + np.random.randn(*y.shape) * noise_std
        else:
            return y       
    else:
        if out_flag == 1:
            return X, y
        else:
            return y

################################# Multi-fidelity ####################################
def calibration_wing(X = None,
    n={'0': 50, '1': 100, '2': 100, '3': 100},
    noise_std={'0': 0.0, '1': 0.0, '2': 0.0, '3': 0.0},
    random_state = None, shuffle = True):
    if X is None:
        X_list = []
        y_list = []
        for level, num in n.items():
            if level  in ['0','1','2','3'] and num > 0:
                X, y = calibration_f_wing(n=num, fidelity = int(level), 
                    noise_std= noise_std[level], random_state = random_state)
                X = np.hstack([X, np.ones(num).reshape(-1, 1) * float(level)])
                X_list.append(X)
                y_list.append(y)
            else:
                raise ValueError('Wrong label, should be h, l1, l2 or l3')
        return np.vstack([*X_list]), np.hstack(y_list)
    else:
        fidelity = [i for i in n.keys()]
        y_list = []
        if type(X) == np.ndarray:
            X = torch.tensor(X)
        for f in fidelity:
            # Find the index for each fidelity then call the wing
            index = [i[0] for i in torch.argwhere(X[...,-1]== int(f))]
            y_list.append(calibration_f_wing(X=X[index, 0:-1], fidelity=int(f),  
                noise_std= noise_std[f]))
        return torch.tensor(np.hstack(y_list))
    


################################## borehole #########################################
def borehole(n=100, X = None, 
    fidelity = 0,noise_std = 0.0, random_state = None, shuffle = True):
    if random_state is not None:
        np.random.seed(random_state)

    dx = 8
    l_bound = [100, 990, 700, 100, .05,6000, 10, 1000]
    u_bound = [1000, 1110, 820, 10000, .15, 12000,500, 2000]
    out_flag = 0
    if X is None:
        sobolset = Sobol(d=dx, seed = random_state)
        X = sobolset.random(2 ** (np.log2(n) + 1).astype(int))[:n, :]
        X = scale(X, l_bounds=l_bound, u_bounds=u_bound)
        out_flag = 1
    if type(X) != np.ndarray:
        X = np.array(X)

    Tu, Hu, Hl, r, rw, Kw,Tl, L = [X[:, i] for i in range(8)]
    # This is the output

    if fidelity == 0:
        Numer = 2*np.pi*Tu
        Tl=250
        L=1500
        X[:, 6]=Tl
        X[:, 7]=L
        y = (Numer*(Hu-Hl))/(np.log(r/rw)*(1+((2*L*Tu)/(np.log(r/rw)*(rw**2)*Kw))+(Tu/Tl)))
    elif fidelity == 1:
    # This is the output
        Numer = 2*np.pi*Tu
        y = (Numer*(Hu-.8*Hl))/(np.log(r/rw)*(1+((1*L*Tu)/(np.log(r/rw)*(rw**2)*Kw))+(Tu/Tl)))

    elif fidelity == 2:
        Numer = 2*np.pi*Tu
    # This is level 2 in Tammers paper
        y = (Numer*(Hu-Hl))/(np.log(r/rw)*(1+((8*L*Tu)/(np.log(r/rw)*(rw**2)*Kw))+.75*(Tu/Tl)))
    else:
        raise ValueError('only 3 fidelities of 0,1, and 2 have been implemented')

    if X is None:
        if shuffle:
            index = np.random.randint(0, len(y), size = len(y))
            X = X[index,...]
            y = y[index]


    if noise_std > 0.0:
        if out_flag == 1:
            return X, y + np.random.randn(*y.shape) * noise_std
        else:
            return y       
    else:
        if out_flag == 1:
            return X, y
        else:
            return y

################################# Calibration_borehole ####################################
def calibration_borehole(X = None,
    n={'0': 50, '1': 100, '2': 100},
    noise_std={'0': 0.0, '1': 0.0, '2': 0.0},
    random_state = None, shuffle = True):
    if X is None:
        X_list = []
        y_list = []
        for level, num in n.items():
            if level  in ['0','1','2'] and num > 0:
                X, y = borehole(n=num, fidelity = int(level), 
                    noise_std= noise_std[level], random_state = random_state)
                X = np.hstack([X, np.ones(num).reshape(-1, 1) * float(level)])
                X_list.append(X)
                y_list.append(y)
            else:
                raise ValueError('Wrong label, should be h, l1 or l2')
        return np.vstack([*X_list]), np.hstack(y_list)
    else:
        fidelity = [i for i in n.keys()]
        y_list = []
        if type(X) == np.ndarray:
            X = torch.tensor(X)
        for f in fidelity:
            # Find the index for each fidelity then call the wing
            index = [i[0] for i in torch.argwhere(X[...,-1]== int(f))]
            y_list.append(borehole(X=X[index, 0:-1], fidelity=int(f),  
                noise_std= noise_std[f]))
        return torch.tensor(np.hstack(y_list))
#####################################################################################################


# Copyright © 2023, Dr. Bostanabad's research group at the University of California, Irvine.
# 
# GP+ Intellectual Property Notice:
# 
# The software known as GP+ is the proprietary material of Dr. Bostanabad's research group at the University of California, Irvine. 
# Non-profit academic institutions and U.S. government agencies may utilize this software exclusively for educational and research endeavors. 
# All other entities are granted permission for evaluation purposes solely; any additional utilization demands prior written consent from the appropriate authority. 
# The direct sale or redistribution of this software, in any form, without explicit written authorization is strictly prohibited. 
# Users are permitted to make duplicate copies of the software, contingent upon the assurance that no copies are sold or redistributed and they adhere to the stipulated terms herein.
# 
# Being academic research software, GP+ is provided on an "as is" m_gp, devoid of warranties, whether explicit or implicit. 
# The act of downloading or executing any segment of this software inherently signifies compliance with these terms. 
# The developers reserve the right to modify these terms and conditions without prior intimation at any juncture.

import torch
import numpy as np
from gpytorch import settings as gptsettings
from gpytorch.utils.errors import NanError,NotPSDError
from scipy.optimize import minimize,OptimizeResult
from collections import OrderedDict
from functools import reduce
from joblib import Parallel,delayed
from joblib.externals.loky import set_loky_pickler
from typing import Dict,List,Tuple,Optional,Union
from copy import deepcopy
from scipy.optimize import Bounds
from scipy.optimize import NonlinearConstraint
from scipy.optimize import BFGS
#######################################################

from gpplus.utils.interval_score import interval_score_function
tkwargs = {
    "dtype": torch.float,
    "device": torch.device("cpu" if torch.cuda.is_available() else "cpu"),
}

def marginal_log_likelihood(model,add_prior:bool,regularization_parameter=[0,0]):
    output = model(*model.train_inputs)
    out = model.likelihood(output).log_prob(model.train_targets)
    if add_prior:
        # add priors
        for _, module, prior, closure, _ in model.named_priors():
            out.add_(prior.log_prob(closure(module)).sum())
    temp = 0
    temp_1=0
    for name, param in model.named_parameters():
        string_list = ['fci', 'h1', 'h2', 'h3', 'h4', 'h5', 'h6', 'h8','h9', 'h10', 'h11', 'h12','fce']
        if name in string_list:
            temp += torch.norm(param)
            temp_1 += torch.sum(torch.abs(param))
        elif name in ['nn_model.' + str + '.bias' for str in string_list]:
            temp += torch.norm(param)
            temp_1 += torch.sum(torch.abs(param))

    out -= regularization_parameter[0]*temp_1 + regularization_parameter[1]* temp    
    ## Interval Score if neede for BO
    if model.interval_score is True:
        score, accuracy = interval_score_function(output.mean + 1.96 * output.variance.sqrt(), output.mean - 1.96 * output.variance.sqrt(), model.y_scaled)
        return out - 0.08*torch.abs(out) * score#- torch.exp(model.interval_alpha) * score
    return out 


class MLLObjective:

    def __init__(self,model,add_prior,regularization_parameter):
        self.model = model 
        self.add_prior = add_prior
        self.regularization_parameter=regularization_parameter

        parameters = OrderedDict([
            (n,p) for n,p in self.model.named_parameters() if p.requires_grad
        ])
        self.param_shapes = OrderedDict()
        for n,p in self.model.named_parameters():
            if p.requires_grad:
                if len(parameters[n].size()) > 0:
                    self.param_shapes[n] = parameters[n].size()
                else:
                    self.param_shapes[n] = torch.Size([1])
    
    def pack_parameters(self) -> np.ndarray:
        parameters = OrderedDict([
            (n,p) for n,p in self.model.named_parameters() if p.requires_grad
        ])
        
        return np.concatenate([parameters[n].cpu().data.numpy().ravel() for n in parameters])
    
    def unpack_parameters(self, x:np.ndarray) -> torch.Tensor:
        i = 0
        named_parameters = OrderedDict()
        for n in self.param_shapes:
            param_len = reduce(lambda x,y: x*y, self.param_shapes[n])
            # slice out a section of this length
            param = x[i:i+param_len]
            # reshape according to this size, and cast to torch
            param = param.reshape(*self.param_shapes[n])
            named_parameters[n] = torch.from_numpy(param).to(**tkwargs)
            # update index
            i += param_len
        return named_parameters

    def pack_grads(self) -> None:
        """Concatenate gradients from the parameters to 1D numpy array
        """
        grads = []
        for name,p in self.model.named_parameters():
            if p.requires_grad:
                grad = p.grad.cpu().data.numpy()
                grads.append(grad.ravel())
        return np.concatenate(grads).astype(np.float64)

    def fun(self, x:np.ndarray,return_grad=True) -> Union[float,Tuple[float,np.ndarray]]:
        # unpack x and load into module 
        state_dict = self.unpack_parameters(x)
        old_dict = self.model.state_dict()
        old_dict.update(state_dict)
        self.model.load_state_dict(old_dict)

        self.model.zero_grad()
        obj = -marginal_log_likelihood(self.model, self.add_prior,self.regularization_parameter) # negative sign to minimize
        
        if return_grad:
            obj.backward()
            
            return obj.item(),self.pack_grads()
        
        return obj.item()


def _sample_from_prior(model) -> np.ndarray:
    out = []
    for _,module,prior,closure,_ in model.named_priors():
        if not closure(module).requires_grad:
            continue
            
        out.append(prior.expand(closure(module).shape).sample().cpu().numpy().ravel())
    
    return np.concatenate(out)


def cons_f(x,likobj):
    zeta = torch.tensor(likobj.model.zeta, dtype = torch.float64)
    A = likobj.unpack_parameters(x)['fci']
    likobj.model.nn_model.fci.weight.data = A
    positions = likobj.model.nn_model(zeta)
    out_constraint=positions.detach().numpy().reshape(-1,)
    return out_constraint[0:8]

def get_bounds(likobj, theta):

    dic = likobj.unpack_parameters(theta)

    minn = np.empty(0)
    maxx = np.empty(0)
    for name, values in dic.items():
        for ii in range(len(likobj.model.qual_kernel_columns)):
            if name ==  str(likobj.model.qual_kernel_columns[ii]):
                minn = np.concatenate( (minn,  np.repeat(-3, values.numel()) ) )
                maxx = np.concatenate( (maxx,  np.repeat(3, values.numel()) ) )
        if name == 'likelihood.noise_covar.raw_noise' or name.startswith('[') or name.startswith('latent['):
            minn = np.concatenate( (minn,  np.repeat(-np.inf, values.numel()) ) )
            maxx = np.concatenate( (maxx,  np.repeat( np.inf, values.numel()) ) )
        if 'raw_lengthscale' in name:
            minn = np.concatenate( (minn,  np.repeat(-10.0, values.numel()) ) )
            maxx = np.concatenate( (maxx,  np.repeat( 3.0, values.numel()) ) )
        elif name.startswith('covar_module'):
            minn = np.concatenate( (minn,  np.repeat(-10.0, values.numel()) ) )
            maxx = np.concatenate( (maxx,  np.repeat( 3.0, values.numel()) ) )
            ######################################################################################### For multiple Bases ##################################
        elif name.startswith('mean'):
            minn = np.concatenate( (minn,  np.repeat(-1.5, values.numel()) ) )
            maxx = np.concatenate( (maxx,  np.repeat( 1.5, values.numel()) ) )
        elif name.startswith('Theta_'):
            minn = np.concatenate( (minn,  np.repeat(-15, values.numel()) ) )
            maxx = np.concatenate( (maxx,  np.repeat( 15, values.numel()) ) )
        elif name.startswith('encoder'):
            minn = np.concatenate( (minn,  np.repeat(-15, values.numel()) ) )
            maxx = np.concatenate( (maxx,  np.repeat( 15, values.numel()) ) )
            ######################################################################################### For A_matrix and Variationa encoder  #################
        elif name.startswith('A_matrix'):
            minn = np.concatenate( (minn,  np.repeat(-10, values.numel()) ) )
            maxx = np.concatenate( (maxx,  np.repeat( 10, values.numel()) ) )
    return np.array(minn).reshape(-1,), np.array(maxx).reshape(-1,)



def _fit_model_from_state(likobj,theta0,jac,options, method = 'trust-constr',constraint=False,bounds=False):
    
    min, max = get_bounds(likobj, theta0)
    bounds_acts = Bounds(min, max)
    nonlinear_constraint = NonlinearConstraint(lambda x: cons_f(x, likobj), [0,0,0,0,-5,0,-5,-5],[0,0,5,0,5,5,5,5], jac='2-point', hess=BFGS())
    '''
    if constraint==True:
        nonlinear_constraint = NonlinearConstraint(lambda x: cons_f(x, likobj), [0,0,0,0,-inf,0],[0,0,inf,0,inf,inf], jac='2-point', hess=BFGS())
    
    else:
        nonlinear_constraint = NonlinearConstraint(lambda x: cons_f(x, likobj), [-inf,-inf,-inf,-inf,-inf,-inf],[inf,inf,inf,inf,inf,inf], jac='2-point', hess=BFGS())

    '''
    eq_cons = {'type': 'eq',
                'fun' : lambda x: np.array([cons_f(x, likobj)[0],cons_f(x, likobj)[1],cons_f(x, likobj)[3]])}
    ineq_cons = {'type': 'ineq',
                'fun' : lambda x: np.array([cons_f(x, likobj)[2],cons_f(x, likobj)[5]])}
    if constraint==True:
        nonlinear_constraint=[nonlinear_constraint]
    else:
        nonlinear_constraint=[]


    if bounds==True:
        bounds=bounds_acts
    else:
        bounds=None


    try:
        with gptsettings.fast_computations(log_prob=False):
            return minimize(
                fun = likobj.fun,
                x0 = theta0,
                args=(True) if jac else (False),

                method = method,
                jac=jac,
                bounds=bounds,
                constraints= nonlinear_constraint,
                #constraints=[eq_cons, ineq_cons],
                options= options 
            )


    except Exception as e:
        if isinstance(e,NotPSDError) or isinstance(e, NanError):
            # Unstable hyperparameter configuration. This can happen if the 
            # initial starting point is bad. 
            return e
        else:
            # There is some other issue, most likely with the inputs supplied
            # by the user. Raise error to indicate the problematic part.
            raise



def fit_model_scipy(
    model,
    add_prior:bool=True,
    num_restarts:int=1,
    theta0_list:Optional[List[np.ndarray]]=None,
    jac:bool=True, 
    options:Dict={},
    n_jobs:int=-1,
    method = 'L-BFGS-B',
    constraint=False,
    bounds=False,
    regularization_parameter:List[int]=[0,0]
    ) -> Tuple[List[OptimizeResult],float]:

    defaults = {}

    if method == 'L-BFGS-B':
        defaults = {'ftol':1e-6,'gtol':1e-5,'maxfun':5000,'maxiter':2000}
    elif method == 'trust-constr':
        defaults = {'verbose': 1}
    elif method == 'BFGS':
        defaults = {'gtol': 1e-07, 'norm': np.inf, 'eps': 1.4901161193847656e-08,'maxiter': None, 'disp': False, 'return_all': False, 'finite_diff_rel_step': None}#
    elif method == 'SLSQP':
        defaults = { 'maxiter': 100, 'ftol': 1e-06, 'iprint': 1, 'disp': False, 'eps': 1.4901161193847656e-08, 'finite_diff_rel_step': None}#
    elif method == 'Newton-CG':
        defaults={'xtol': 1e-05, 'eps': 1.4901161193847656e-08, 'maxiter': None, 'disp': False, 'return_all': False}
    else:
        raise ValueError('Wrong method')

    if len(options) > 0:
        for key in options.keys():
            if key not in defaults.keys():
                raise RuntimeError('Unknown option %s!'%key)
            defaults[key] = options[key]

    likobj = MLLObjective(model,add_prior,regularization_parameter)

    if theta0_list is None:
        theta0_list = [likobj.pack_parameters()]
        if num_restarts > -1:
            theta0_list.extend([_sample_from_prior(model) for _ in range(num_restarts+1)])
            theta0_list.pop(0)                                                                     
    
    set_loky_pickler("dill") 

    out = Parallel(n_jobs=n_jobs,verbose=0)(
        delayed(_fit_model_from_state)(likobj,theta0,jac,defaults, method,constraint,bounds) \
            for theta0 in theta0_list
    )
    set_loky_pickler("pickle")

    nlls_opt = [np.inf if isinstance(res,Exception) else res.fun for res in out]
    best_idx = np.argmin(nlls_opt)
    try:
        theta_best = out[best_idx].x
        old_dict = deepcopy(model.state_dict())
        old_dict.update(likobj.unpack_parameters(theta_best))
        model.load_state_dict(old_dict)

        if 'fci' in [name for name,p in model.named_parameters()]:
            model.nn_model.fci.weight.data = old_dict['fci']       
    except:
        pass
    return out,nlls_opt[best_idx]


import numpy as np
import torch
from gpplus.models import GP_Plus
from gpplus.optim import fit_model_scipy
from joblib import Parallel, delayed
from joblib.externals.loky import set_loky_pickler
from gpplus.bayesian_optimizations.AFs import AF_HF,AF_LF, AF_HF_Engineering,AF_LF_Engineering
import matplotlib.pyplot as plt

    
def BO(Xtrain=None,
        ytrain=None,
        costs=None,
        l_bound=None,
        u_bound=None,
        xmean=None,
        xstd=None,
        qual_index=None,
        data_gen_func=None,
        n_train=None,
        maximize_flag=False,
        one_iter=False,
        max_cost=40000,
        MF=True,
        AF_hf=AF_HF,
        AF_lf=AF_LF,
        max_iter=2,
        IS=True):  
      
    ymin_list = []
    xmin_list = []   
    cumulative_cost = []
    bestf = []
    Fidelity=[]
    num_fidelity=list(qual_index.values())[-1]
    def cost_fun(x):
        return costs[str(int(x))]
    
    def bestf_calculator(MF,num_fidelity,ytrain,Xtrain):
        if MF:
            best_values = []
            if maximize_flag:
                for i in range(num_fidelity):
                    best_values.append((ytrain[Xtrain[:,-1] == i].max().reshape(-1,).item()))
            else:
                for i in range(num_fidelity):
                    best_values.append((ytrain[Xtrain[:,-1] == i].min().reshape(-1,).item()))
        else:
            best_values = []
            if maximize_flag:
                best_values.append((ytrain.max().reshape(-1,)).item())
            else:
                best_values.append((ytrain.min().reshape(-1,)).item())
        
        return best_values
    
    
    def run_scipy(EI, best_f, bound, model,xmean,xstd,cost_fun, fidelity,l_bound,u_bound):
        import torch
        import numpy as np
        from scipy.optimize import minimize
        random_seed = np.random.choice(range(0,1000), size=12, replace=False)
        def run(EI, best_f, bound, model,xmean,xstd,cost_fun,k, fidelity,l_bound_h,u_bound_h):
            l_bound_h=l_bound+[fidelity]
            u_bound_h=u_bound+[fidelity]
            np.random.seed(random_seed[k])
            samples_h = torch.tensor((np.random.uniform(l_bound_h,u_bound_h)).reshape(1,-1))
            samples_h[:,-1] = torch.round(samples_h[:,-1])
            result_h = minimize(EI,samples_h.reshape(-1,), args=(best_f, model,xmean,xstd,cost_fun),bounds=bound)
            temp = result_h.fun
            tempx = result_h.x
            return temp, tempx
        
        set_loky_pickler("dill") 
        out = Parallel(n_jobs=-1,verbose=0)(delayed(run)(EI, best_f, bound, model, xmean,xstd,cost_fun,k,fidelity,l_bound,u_bound) for k in range(12))
        set_loky_pickler("pickle")
        temp = [out[i][0] for i in range(len(out))]
        tempx = [out[i][1] for i in range(len(out))]
        min_index = np.argmin(temp)
        Y_h = temp[min_index]
        X_h = tempx[min_index]
        return Y_h, X_h
    

    
    if callable(data_gen_func):
        initial_cost = np.sum(list(map(cost_fun,Xtrain[:,-1])))
        cumulative_cost.append(initial_cost)
        
        
        if type(ytrain)!=torch.tensor:
            ytrain=torch.tensor(ytrain)
        ytrain=ytrain.reshape(-1)
        problem = lambda x: data_gen_func(False,x)
        while cumulative_cost[-1] < max_cost:
            
            best_values=bestf_calculator(MF,num_fidelity,ytrain,Xtrain)
            bestf.append(best_values[0])
            if len(bestf)> max_iter:
                if np.var(bestf[-max_iter:])<1e-6:
                    break
                
            model = GP_Plus(Xtrain, ytrain, qual_index,IS=IS)

            _ = fit_model_scipy(model, bounds=True)
            
                
            X_list=[]   
            y_list=[]
            
            for i in range(num_fidelity):
                if i==0:
                    bound=tuple(list(zip(l_bound, u_bound))+[(i,i)])
                    Y_h, X_h = run_scipy(AF_hf, best_values[0], bound, model,np.array(xmean),np.array(xstd),cost_fun, i,l_bound,u_bound)
                    X_list.append(X_h)
                    y_list.append(Y_h)
                else:
                    bound=tuple(list(zip(l_bound, u_bound))+[(i,i)])
                    Y_l, X_l = run_scipy(AF_lf, best_values[0], bound, model,np.array(xmean),np.array(xstd),cost_fun, i,l_bound,u_bound)
                    X_list.append(X_l)
                    y_list.append(Y_l)
            
                
                    
            min_index = np.argmin(y_list)
            temp = torch.tensor(X_list[min_index])
            ynew = problem(temp.unsqueeze(0))
            
            
            if MF:
                Xnew= np.concatenate([((temp[0:-1]-xmean)/xstd).reshape(1,-1), temp[-1].reshape(-1,1)],axis=-1)
            else:
                Xnew= ((temp-xmean)/xstd).reshape(1,-1)
                

                
            
            Xtrain = torch.cat([Xtrain,torch.tensor(Xnew.reshape(1,-1))])
            ytrain = torch.cat([ytrain, ynew.reshape(-1,)])
            ymin_list.append(ynew.reshape(-1,))
            xmin_list.append(Xnew)
            ##################
            cumulative_cost.append(initial_cost + cost_fun(Xnew[0][-1]))
            initial_cost = cumulative_cost[-1]
            Fidelity.append(Xnew[0][-1])
            if one_iter:
                best_values=bestf_calculator(MF,num_fidelity,ytrain,Xtrain)
                bestf.append(best_values[0])
                break

        
    else:
        Xtrain = np.empty((0, data_gen_func.shape[1]-1))
        ytrain = np.empty((0,))
        for i in range(num_fidelity):
            random_index=np.random.randint(0, len(data_gen_func[data_gen_func[:,-2]==i]), n_train[i])
            Xtrain = np.append(Xtrain, data_gen_func[data_gen_func[:,-2]==i][random_index][:,0:-1], axis=0)
            ytrain = np.append(ytrain, data_gen_func[random_index][:,-1], axis=0)
            
            
        Xtrain=torch.tensor(Xtrain)
        ytrain=torch.tensor(ytrain)
        initial_cost = np.sum(list(map(cost_fun,Xtrain[:,-1])))
        cumulative_cost.append(initial_cost)
        
        
        while cumulative_cost[-1] < max_cost:
            best_values=bestf_calculator(MF,num_fidelity,ytrain,Xtrain)
                    
            bestf.append(best_values[0])
            if len(bestf)> max_iter:
                if np.var(bestf[-max_iter:])<1e-6:
                    break
                
            model = GP_Plus(Xtrain, ytrain, qual_index,IS=IS)

            _ = fit_model_scipy(model, bounds=True)
            

            X_list=[]   
            y_list=[]
            
            for i in range(num_fidelity):
                scores=[]
                with torch.no_grad():
                    ytest, ystd = model.predict(torch.tensor(data_gen_func[data_gen_func[:,-2]==i][:,0:-1]), return_std=True,include_noise = False)
                if i==0:
                    # scores.append(AF_hf(best_values[i], ytest.reshape(-1,1), ystd.reshape(-1,1), maximize = maximize_flag)) 
                    scores.append(AF_HF_Engineering(best_values[i], ytest.reshape(-1,1), ystd.reshape(-1,1),torch.tensor(data_gen_func[data_gen_func[:,-2]==i][:,0:-1]),cost_fun,maximize = maximize_flag)) 
                else:
                    scores.append(AF_LF_Engineering(best_values[i], ytest.reshape(-1,1), ystd.reshape(-1,1),torch.tensor(data_gen_func[data_gen_func[:,-2]==i][:,0:-1]),cost_fun,maximize = maximize_flag))
            
            scores_final=torch.cat(scores,dim=0)
            index = torch.argmax(scores_final)
            
            Xnew = data_gen_func[index][0:-1]
            Xnew=torch.tensor(Xnew)
            ynew = data_gen_func[index][-1]
            
            Xtrain = torch.cat([Xtrain,Xnew.reshape(1,-1)])
            ytrain = torch.cat([ytrain, torch.tensor(ynew).reshape(-1,)], dim=0)
            ymin_list.append(ynew.reshape(-1,))
            xmin_list.append(Xnew)

            cumulative_cost.append(initial_cost + cost_fun(Xnew[-1]))
            initial_cost = cumulative_cost[-1]
            Fidelity.append(Xnew[-1])
            if one_iter:
                best_values=bestf_calculator(MF,num_fidelity,ytrain,Xtrain)
                bestf.append(best_values[0])
                break
    

    

    return np.array(bestf), np.array(cumulative_cost)



def Visualize_BO(bestf,cost):
    plt.scatter(cost,bestf)
    plt.ylabel('y^*')
    plt.xlabel('cost')
    plt.show()
    






import torch
import numpy as np
from gpplus.utils import data_type_check

def setlevels(X, qual_index = None, return_label = False):
    labels = []
    if qual_index == []:
        return X
    if qual_index is None:
        qual_index = list(range(X.shape[-1]))
    X=data_type_check(X)
    # Check if X is a PyTorch tensor
    if isinstance(X, torch.Tensor):
        # Move X to CPU if it's on CUDA
        if X.is_cuda:
            X = X.cpu()
        temp = X.clone()
    # Check if X is a NumPy array
    else:
        # Handle other types or raise an error
        raise TypeError("X must be a PyTorch tensor or a NumPy array.")

    # Convert PyTorch tensor to NumPy array if needed
    if isinstance(temp, torch.Tensor):
        temp = temp.numpy()

    if temp.ndim > 1:
        for j in qual_index:
            l = np.sort(np.unique(temp[..., j])).tolist()
            labels.append(l)
            #l =  torch.unique(temp[..., j], sorted = True).tolist()
            temp[..., j] = torch.tensor([*map(lambda m: l.index(m),temp[..., j])])
    else:
            l = torch.unique(temp, sorted = True)
            temp = torch.tensor([*map(lambda m: l.tolist().index(m), temp)])
    
    
    if temp.dtype == object:
        temp = temp.astype(float)
        if type(X) == np.ndarray:
            temp = torch.from_numpy(temp)
        
        if return_label:
            return temp, labels
        else:
            return temp
    else:
        if type(X) == np.ndarray:
            temp = torch.from_numpy(temp)
        if return_label:
            return temp, labels
        else:
            return temp


# Copyright © 2023, Dr. Bostanabad's research group at the University of California, Irvine.
# 
# GP+ Intellectual Property Notice:
# 
# The software known as GP+ is the proprietary material of Dr. Bostanabad's research group at the University of California, Irvine. 
# Non-profit academic institutions and U.S. government agencies may utilize this software exclusively for educational and research endeavors. 
# All other entities are granted permission for evaluation purposes solely; any additional utilization demands prior written consent from the appropriate authority. 
# The direct sale or redistribution of this software, in any form, without explicit written authorization is strictly prohibited. 
# Users are permitted to make duplicate copies of the software, contingent upon the assurance that no copies are sold or redistributed and they adhere to the stipulated terms herein.
# 
# Being academic research software, GP+ is provided on an "as is" m_gp, devoid of warranties, whether explicit or implicit. 
# The act of downloading or executing any segment of this software inherently signifies compliance with these terms. 
# The developers reserve the right to modify these terms and conditions without prior intimation at any juncture.


import math
import numpy as np
import torch
from warnings import simplefilter,catch_warnings
from gpytorch import settings as gptsettings
from gpytorch.utils.warnings import GPInputWarning
from scipy.spatial import distance_matrix
from .mll_scipy import fit_model_scipy
from ..models import GPR
from copy import deepcopy
from typing import Tuple,Dict

def loocv_rrmse(model:GPR):
    # generate prediction caches (using predictions)
    model.eval()
    if model.prediction_strategy is None:
        with torch.no_grad(), gptsettings.fast_computations(log_prob=False):
            with catch_warnings():
                simplefilter('ignore',category=GPInputWarning)
                _ = model(*model.train_inputs).mean

    Kinv_y = model.prediction_strategy.mean_cache
    Rinv = model.prediction_strategy.covar_cache # RinvRinv^T= Kinv
    Kinv_diag = (Rinv**2).sum(dim=-1)

    loo_error = Kinv_y/Kinv_diag
    return (loo_error**2).mean().sqrt().item()

# NLL is negative loglikelihood
def fit_model_continuation(
    model:GPR,add_prior:bool=True,
    num_restarts:int=32,criterion:str='NLL',
    initial_noise_var:float=1,
    red_factor:float=math.sqrt(10),
    options:Dict={},
    n_jobs:int= -1,
    accuracy = 1e-2,
    method = 'L-BFGS-B',
    constraint=False,
    regularization_parameter=[0, 0],
    bounds=False
)-> Tuple[float,Dict]:
    """Sequentially optimize the log-likelihood of a standard GP model for a decreasing
    sequence of noise variances.

    This function, based on the work of `Bostanabad et al. (2018)`_, leverages the smoothing 
    effect of the noise variance on the log-likelihood profile. At each iteration, the function 
    calls `fit_model_scipy` to optimize the log-likelihood while holding the noise variance fixed. 
    The noise variance is then halved for the next iteration. The iteration is terminated when one
    of the following occurs:
        1. criterion begins to increase
        2. a lower bound on the noise variance is reached
        3. cholesky matrix is singular

    If `num_restarts` > 0, multistart optimization is used at each step. In the first iteration,
    the current state of the model and `num_restarts` samples drawn from the prior distribution are 
    used as starting points. In each subsequent iteration, (distinct) optima from the previous iteration
    are used as starting points.
    
    Unlike `Bostanabad et al. (2018)`_, this function uses negative log-likelihood in place of leave
    one-out cross-validation (LOOCV) RMSE as one of the termination criterion. We find that LOOCV is
    unreliable in the case of LVGPs. 

    The initial noise variance should be specifed before passing the model. A value of 1 works for
    many situations. A higher value may be needed when there are many observations and many 
    hyperparameters to estimate. 

    :param model: A model instance derived from the `models.GPR` class. Can also pass a instance
        inherting from `gpytorch.models.ExactGP` provided that `num_restarts=0` or 
        the class implements a `.reset_parameters` method.
    :type model: models.GPR

    :param add_prior: Whether to add the hyperparameter priors to the log-likelihood to optimize the 
        posterior. Optimizing the log-posterior is some what more robust than optimizing the log-likelihood
        when there are few training data. Defaults to True
    :type num_restarts: bool, optional

    :param num_restarts: The number of times to restart the local optimization from a new starting 
        point at the initial noise variance. Subsequent optimizations steps are initialized from
        from the (distinct) optima at the previous step. Defaults to 5.
    :type num_restarts: int, optional

    :param criterion: The criterion used for termination. Can either be 'NLL' or 'LOOCV'. Defaults
        to 'NLL'
    :type criterion: str, optional

    :param initial_noise_var: The initial noise variance. Defaults to 1.
    :type initial_noise_var: float, optional

    :param red_factor: Factor to reduce the noise variance by at each step. This needs to be > 2. 
        Defaults to sqrt(10).
    :type red_factor: float,optional

    :param options: A dictionary of `L-BFGS-B` options to be passed to `scipy.optimize.minimize`.
    :type options: dict,optional

    :param n_jobs: Number of jobs to run in parallel. Uses `joblib` for parallelization. Deafults to 1. 
    :type n_jobs: int,optional

    .. _Bostanabad et al. (2018):
        https://doi.org/10.1002/nme.5751

    Returns:
        A two element tuple with the following elements
            - the value of the criterion at termination
            - a dictionary with the following entries
                - 'noise_history': list of noise variances tried
                - 'nll_history': negative log-likelihoods of the optimal hyperparameters at each noise variance
                - 'loocv_history': LOOCV RRMSEs of the optimal hyperparameters at each noise variance
                - 'optimization_history': list of list of optimization result objects
    """
    if criterion.upper() not in ['NLL','LOOCV']:
        raise AttributeError('criterion must be one of NLL or LOOCV')

    if red_factor < 2:
        raise RuntimeError('Reduction factor for noise variance needs to be greater then 2')

    if model.likelihood.raw_noise.requires_grad:
        model.likelihood.raw_noise.requires_grad_(False)
    
    t = 0    
    theta0_list = None
    while True:

        t += 1
        initial_noise_var_new = initial_noise_var
        if t == 1:
            noises = [initial_noise_var_new/(10**i) for i in range(int(10/t))]
            # noises = [initial_noise_var_new/(np.sqrt(10)**i) for i in range(int(100/t))]

        else:
            if index >= 2 and index < len(history['noise_history'])-2:
                noises = np.linspace(history['noise_history'][index-1], history['noise_history'][index+1], 10)
                initial_noise_var = history['noise_history'][index-1]
                model.load_state_dict(old_state_dict[index-1])
            elif index >= 1 and index < len(history['noise_history'])-1:
                noises = np.linspace(history['noise_history'][index-1], history['noise_history'][index+1], 10)
                initial_noise_var = history['noise_history'][index-1]
                model.load_state_dict(old_state_dict[index-1])
            else:
                model.load_state_dict(old_state_dict[index])
                nllll=history['nll_history'][index]
                print(f'Negative log likelihood={nllll}')
                return history['nll_history'][index],history
                

        noise_list = []
        nll_list = []
        loocv_list = []
        reslist_list = []
        t += 1
        old_crit = math.inf

        old_state_dict = {}


        for i in range(len(noises)):
            model.train()
            model.likelihood.initialize(**{'noise':noises[i]})
            old_state_dict[i] = deepcopy(model.state_dict())
            
            reslist,nll = fit_model_scipy(
                model,add_prior,num_restarts=num_restarts,theta0_list=theta0_list,options=options, n_jobs= n_jobs, method=method, constraint=constraint, regularization_parameter=regularization_parameter, bounds=bounds
            )
            
            if all([isinstance(res,RuntimeError) or isinstance(res,TypeError) for res in reslist]):
                # some error
                break
            
            noise_list.append(model.likelihood.noise.data)
            nll_list.append(nll)
            loocv_list.append('NLL') ## Amin Changend loocv_list.append(loocv_rrmse(model))
            reslist_list.append(reslist)
            crit = nll if criterion.upper()=='NLL' else loocv_list[-1]   #Amin changed
            
            '''
            if crit >= old_crit:
                model.load_state_dict(old_state_dict)
                best_crit = old_crit
                if red_factor > 2:
                    red_factor = 2
                    model.likelihood.initialize(**{'noise':noise_list[-2]/red_factor})
                    continue
                break
            '''

            theta0_list = []
            for res in reslist:
                if isinstance(res,Exception):
                    continue
                if len(theta0_list) > 0:
                    dists = distance_matrix(res.x.reshape(1,-1),np.row_stack(theta0_list)).ravel()
                    if np.any(dists < 1e-2*res.x.shape[0]):
                        continue

                theta0_list.append(res.x)
            
            try:
                model.likelihood.initialize(**{'noise':noise_list[-1]/red_factor})
            except:
                # violates constraints
                try:
                    model.likelihood.initialize(**{'noise':noise_list[-1]/red_factor+1e-10})
                except:
                    break
            old_crit =  math.inf #crit   Amin changed

        history = {
        'noise_history':noise_list,
        'nll_history':nll_list,
        'loocv_history':loocv_list,
        'optimization_history':reslist_list
        }

        index = np.argmin(history['nll_history'])


        print('Finished for loop')


        print(history['nll_history'])

        if np.abs(initial_noise_var_new - history['noise_history'][index]) < accuracy:
            model.load_state_dict(old_state_dict[index])
            break

        

    return history['nll_history'][index],history



from gpytorch.kernels import RBFKernel
from gpytorch.functions import RBFCovariance
from gpytorch.settings import trace_mode
import torch

def postprocess_rbf(dist_mat):
    return dist_mat.div_(-1).exp_()


class wighted_RBF_Z(RBFKernel):
    """_summary_

    Args:
        RBFKernel (_type_): _description_
    """
    has_lengthscale = True

    def forward(self, x1, x2, diag=False, **params):
        # weight_value=torch.min(torch.tensor(1/self.lengthscale),torch.tensor(1)).item()
        weight_value=1/self.lengthscale
        # weight_value=1
        # print(f'weight_value ={weight_value}')

        if (
            x1.requires_grad
            or x2.requires_grad
            or (self.ard_num_dims is not None and self.ard_num_dims > 1)
            or diag
            or params.get("last_dim_is_batch", False)
            or trace_mode.on()
        ):

            ten_power_omega_sqrt = self.lengthscale.sqrt()
            
            # weight_value=self.lengthscale
            # x1=x1[16,:]* torch.ones_like(x1)
            # x1_ = x1[:,1:].mul(ten_power_omega_sqrt)
            # #### Amin Added this
            # x2_ = x2[:,1:].mul(ten_power_omega_sqrt)


            # return 0*torch.eye(COV_old.size()[0])
            # return self.covar_dist(
            #     x1_, x2_, square_dist=True, diag=diag, dist_postprocess_func=postprocess_rbf, postprocess=True, **params
            # )
        # return RBFCovariance.apply(
        #     x1,
        #     x2,
        #     self.lengthscale,
        #     lambda x1, x2: self.covar_dist(
        #         x1, x2, square_dist=True, diag=False, dist_postprocess_func=postprocess_rbf, postprocess=False, **params
        #     ),
        # )
            x1_ = x1[:,1:].div(self.lengthscale)
            x2_ = x2[:,1:].div(self.lengthscale)
            x_INDEC=x1[:,0]
            COV_old=self.covar_dist(
                x1_, x2_, square_dist=True, diag=diag, dist_postprocess_func=postprocess_rbf, postprocess=True, **params
            )
            W=(x_INDEC==0).nonzero()
            H_value=x2_[W[0]]
            H_value_vectro=H_value*torch.ones_like(x2_)
            # COV_new=torch.diag(torch.norm(x1_-H_value, dim=1))

            Temp=self.covar_dist(
                x1_, H_value_vectro, square_dist=True, diag=diag, dist_postprocess_func=postprocess_rbf, postprocess=True, **params
            )
            COV_new=torch.diag(torch.diagonal(Temp, 0))


            
            weight=torch.ones_like(COV_old)
            weight[W,:]=weight[W,:]/weight_value
            weight[:,W]=weight[:,W]/weight_value

            # weight[W,W]=weight_value*weight[W,W]
            # return  torch.mul(COV_old,weight_value*weight)
            return  weight_value*weight
        else:
            COV_old=RBFCovariance.apply(
                x1,
                x2,
                self.lengthscale,
                lambda x1, x2: self.covar_dist(
                    x1[:,1:], x2[:,1:], square_dist=True, diag=False, dist_postprocess_func=postprocess_rbf, postprocess=False, **params
                ),
            )
            x_INDEC_1=x1[:,0]
            x_INDEC_2=x2[:,0]
            W_1=(x_INDEC_1==0).nonzero()
            W_2=(x_INDEC_2==0).nonzero()

            # weight_value=self.lengthscale
            
            weight=torch.ones_like(COV_old)
            # for i in W_2:
            #     weight[W_1,i]=weight[W_1,i]/weight_value
            weight[W_1,:]=weight[W_1,:]/weight_value
            weight[:,W_2]=weight[:,W_2]/weight_value

            # for i in W_2:
            #     weight[W_1,i]=weight[W_1,i]*weight_value

            # return torch.mul(COV_old,weight_value*weight)
            
            return weight_value*weight

        #     return torch.ones_like(COV_old)
        
        # return torch.ones_like(RBFCovariance.apply(
        #     x1,
        #     x2,
        #     self.lengthscale,
        #     lambda x1, x2: self.covar_dist(
        #         x1, x2, square_dist=True, diag=False, dist_postprocess_func=postprocess_rbf, postprocess=False, **params
        #     ),
        # ))
    


        #     return torch.eye(torch.zeros_like(COV_old).size()[0])
        
        # return torch.eye(torch.zeros_like(RBFCovariance.apply(
        #     x1,
        #     x2,
        #     self.lengthscale,
        #     lambda x1, x2: self.covar_dist(
        #         x1, x2, square_dist=True, diag=False, dist_postprocess_func=postprocess_rbf, postprocess=False, **params
        #     ),
        # )).size()[0])
        
        # return RBFCovariance.apply(
        #     x1,
        #     x2,
        #     self.lengthscale,
        #     lambda x1, x2: self.covar_dist(
        #         x1, x2, square_dist=True, diag=False, dist_postprocess_func=postprocess_rbf, postprocess=False, **params
        #     ),
        # )


# Copyright © 2021 by Northwestern University.
# 
# LVGP-PyTorch is copyrighted by Northwestern University. It may be freely used 
# for educational and research purposes by  non-profit institutions and US government 
# agencies only. All other organizations may use LVGP-PyTorch for evaluation purposes 
# only, and any further uses will require prior written approval. This software may 
# not be sold or redistributed without prior written approval. Copies of the software 
# may be made by a user provided that copies are not sold or distributed, and provided 
# that copies are used under the same terms and conditions as agreed to in this 
# paragraph.
# 
# As research software, this code is provided on an "as is'' basis without warranty of 
# any kind, either expressed or implied. The downloading, or executing any part of this 
# software constitutes an implicit agreement to these terms. These terms and conditions 
# are subject to change at any time without prior notice.

import math
import torch
from gpytorch.priors import Prior
from torch.distributions import HalfCauchy,HalfNormal,constraints
from torch.distributions.utils import broadcast_all
from numbers import Number

class LogHalfHorseshoePrior(Prior):
    """Prior for the log-noise variance hyperparameter for GPs. 

    This is parameterized by `scale` and `lb`. `lb` is the lower bound on the noise variance. 
    The `scale` parameter is more important. The default value for `scale` - 0.01 - works well 
    for deterministic and low-noise situations. A larger value may be need in noisy sitations 
    and small training datasets. A larger scale implies more noisy data as the prior.

    To change the scale for this prior for a model to say 0.1,
        >>> model.likelihood.register(
        >>>     'noise_prior',LogHalfHorseshoePrior(0.1,model.likelihood.noise_prior.lb),
        >>>     'raw_noise'
        >>> )

    .. note::
        The `log_prob` method is only approximate and unnormalized. There is no closed form
        expression for the underlying horseshoe distribution. The lower and upper bounds on
        its' density are, however, known. Here, we use the same approximate density value that 
        the spearmint package uses.
    
    :param scale: scale parameter of the Horseshoe distribution
    :type scale: float or torch.Tensor

    :param lb: lower bound on the original scale. Defaults to 1e-6
    :type lb: float or torch.Tensor, optional
    """
    arg_constraints = {"scale": constraints.positive,"lb":constraints.positive}
    support = constraints.real
    def __init__(self, scale, lb=1e-6,validate_args=None):
        self.scale,self.lb = broadcast_all(scale,lb)
        if isinstance(scale,Number):
            batch_shape = torch.Size()
        else:
            batch_shape = self.scale.size()
        super().__init__(batch_shape,validate_args=validate_args)
    
    def transform(self, x):
        return self.lb + torch.exp(x)

    def log_prob(self, X):
        # first term is the density in the original scale
        # the second term is for the transformation
        return torch.log(torch.log(1+3*(self.scale / self.transform(X)) ** 2))+ X

    def rsample(self, sample_shape=torch.Size([])):
        local_shrinkage = HalfCauchy(1).rsample(self.scale.shape).to(self.lb)
        param_sample = HalfNormal(local_shrinkage * self.scale).rsample(sample_shape).to(self.lb)
        if len(self.lb) > 1:
            param_sample[param_sample<self.lb[0]] = self.lb[0]
        else:
            param_sample[param_sample<self.lb] = self.lb
        return param_sample.log()

    def expand(self,expand_shape, _instance=None):
        batch_shape = torch.Size(expand_shape)
        return LogHalfHorseshoePrior(self.scale.expand(batch_shape))


from .matern import Matern32Kernel,Matern52Kernel
from gpytorch.kernels import ScaleKernel,RBFKernel,LinearKernel
from .Rough_RBF import Rough_RBF
#### Amin Added  To Test wighted_RBF ####
from .wighted_RBF import wighted_RBF
from .wighted_RBF_Z import wighted_RBF_Z



def AF_LF(samples,best_f,model,xmean,xstd,cost_fun, maximize = False, si = 0.0):
    import numpy as np
    import torch
    from torch.distributions import Normal
    samples= np.concatenate([((samples[0:-1]-xmean)/xstd).reshape(1,-1), samples[-1].reshape(-1,1)], axis = -1)
    samples=torch.tensor(samples.reshape(1,-1))
    with torch.no_grad():
        mean, std = model.predict(samples, return_std=True, include_noise = True)

    cost = torch.tensor(list(map(cost_fun, samples[:,-1].clone().detach())))

    mean=mean.reshape(-1,1)
    view_shape = mean.shape[:-2] if mean.shape[-2] == 1 else mean.shape[:-1]
    mean = mean.view(view_shape)
    sigma = std.view(view_shape)
    u = (mean - best_f - np.sign(best_f) * si) / sigma
    
    if cost is None:
        cost = torch.ones(u.shape)    

    cost = cost.view(u.shape)

    if not maximize:
        u = -u
        
    
    normal = Normal(torch.zeros_like(u), torch.ones_like(u))
    ucdf = normal.cdf(u)
    updf = torch.exp(normal.log_prob(u))
    ei = sigma * updf
    return -1 * (ei/cost)
        
        
def AF_HF(samples,best_f, model,xmean,xstd,cost_fun, maximize = False, si = 0.0,data_gen_func=None):
    import numpy as np
    import torch
    from torch.distributions import Normal
    
    samples= np.concatenate([((samples[0:-1]-xmean)/xstd).reshape(1,-1), samples[-1].reshape(-1,1)], axis = -1)
    samples=torch.tensor(samples.reshape(1,-1))
    with torch.no_grad():
        mean, std = model.predict(samples, return_std=True, include_noise = True)

    mean=mean.reshape(-1,1)

    cost = torch.tensor(list(map(cost_fun, samples[:,-1].clone().detach())))

    # deal with batch evaluation and broadcasting
    view_shape = mean.shape[:-2] if mean.shape[-2] == 1 else mean.shape[:-1]
    mean = mean.view(view_shape)
    sigma = std.view(view_shape)
    u = (mean - best_f - np.sign(best_f) * si) / sigma
    
    if cost is None:
        cost = torch.ones(u.shape)    

    cost = cost.view(u.shape)

    if not maximize:
        u = -u

    ei = sigma * u

    return -1* (ei/cost)
        
        
def AF_EI(samples,best_f, model,xmean,xstd,cost_fun, maximize = False, si = 0.0):
    import numpy as np
    import torch
    from torch.distributions import Normal
    samples= np.concatenate([((samples[0:-1]-xmean)/xstd).reshape(1,-1), samples[-1].reshape(-1,1)], axis = -1)
    samples=torch.tensor(samples.reshape(1,-1))
    with torch.no_grad():
        mean, std = model.predict(samples, return_std=True, include_noise = True)

    mean=mean.reshape(-1,1)

    cost = torch.tensor(list(map(cost_fun, torch.tensor(samples[:,-1].clone().detach(), dtype = torch.int64))))

    # deal with batch evaluation and broadcasting
    view_shape = mean.shape[:-2] if mean.shape[-2] == 1 else mean.shape[:-1]
    mean = mean.view(view_shape)
    sigma = std.view(view_shape)
    u = (mean - best_f - np.sign(best_f) * si) / sigma
    
    if cost is None:
        cost = torch.ones(u.shape)    

    cost = cost.view(u.shape)

    if not maximize:
        u = -u
        
    
    normal = Normal(torch.zeros_like(u), torch.ones_like(u))
    ucdf = normal.cdf(u)
    updf = torch.exp(normal.log_prob(u))
    ei= sigma * (updf + u * ucdf)
    return -1* (ei/cost)
        
        
def AF_LF_Engineering(best_f, mean, std,x_val,cost_fun, maximize = True, si = 0.0, cost = None):
    import numpy as np
    import torch
    from torch.distributions import Normal
    
    view_shape = mean.shape[:-2] if mean.shape[-2] == 1 else mean.shape[:-1]
    mean = mean.view(view_shape)
    sigma = std.view(view_shape)
    
    cost = torch.tensor(list(map(cost_fun, x_val[:,-1].clone().detach())))


    
    u = (mean - best_f - np.sign(best_f) * si) / sigma
    
    if cost is None:
        cost = torch.ones(u.shape)    

    cost = cost.view(u.shape)

    if not maximize:
        u = -u
        #si = -si
    
    normal = Normal(torch.zeros_like(u), torch.ones_like(u))
    ucdf = normal.cdf(u)
    updf = torch.exp(normal.log_prob(u))
    ei = sigma * updf
    return (ei/cost)


def AF_HF_Engineering(best_f, mean, std,x_val, cost_fun,maximize = True, si = 0.0):
    import numpy as np
    import torch
    from torch.distributions import Normal
    view_shape = mean.shape[:-2] if mean.shape[-2] == 1 else mean.shape[:-1]
    mean = mean.view(view_shape)
    sigma = std.view(view_shape)



    cost = torch.tensor(list(map(cost_fun, x_val[:,-1].clone().detach())))

    # deal with batch evaluation and broadcasting
    
    u = (mean - best_f - np.sign(best_f) * si) / sigma
    
    if cost is None:
        cost = torch.ones(u.shape)    

    cost = cost.view(u.shape)

    if not maximize:
        u = -u
        #si = -si
    
    normal = Normal(torch.zeros_like(u), torch.ones_like(u))
    ei = sigma * u
    return (ei/cost)


from setuptools import setup, find_packages

setup(
    name='gpplus',
    version='0.0.2.7.1',
    author='Amin Yousefpour, Ramin Bostanabad',
    author_email='yousefpo@uci.edu',
    description='Python Library for Generalized Gaussian Process Modeling',
    # long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/Bostanabad-Research-Group/GP-Plus',
    packages=find_packages(),
    install_requires=[
        'numpy==1.24.2',
        'scipy==1.11.1',
        'gpytorch==1.7.0',
        'matplotlib==3.7.1',
        'sobol_seq', 
        'tabulate==0.9.0',
        'pandas==1.5.2',
        'dill ==0.3.7',
        'pyro-ppl==1.8.0',
        'pyDOE',
        'botorch==0.6.4',
    ],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.9',
)




import numpy as np
import torch
import warnings



def standard(Xtrain, qual_index, Xtest = None):
    quant_index=list(range(Xtrain.shape[1]))
    for item in qual_index.keys():
        quant_index.remove(item)

    if not isinstance(Xtrain, torch.Tensor):
        original_type = type(Xtrain).__name__
        Xtrain = torch.tensor(Xtrain)

    if Xtest is not None and not isinstance(Xtest, torch.Tensor):
        original_type = type(Xtest).__name__
        Xtest = torch.tensor(Xtest)
    
    
    if len(quant_index) == 0:
        return Xtrain
    
    temp = Xtrain[..., quant_index]
    if type(temp) != torch.Tensor:
        temp = temp.astype(float)

    # mean_xtrain = temp.mean(axis = 0)
    # std_xtrain = temp.std(axis = 0)
    mean_xtrain, std_xtrain = compute_mean_std(temp)
    # Check for NaN values in the original data
    if torch.isnan(temp).any():
        print("Warning: There are NaN values in the data. Mean and standard deviation were calculated excluding these values.")

    temp=(temp - mean_xtrain)/std_xtrain
    Xtrain[..., quant_index] = temp
    if type(Xtrain) == np.ndarray:
        Xtrain = torch.from_numpy(Xtrain)
    if Xtest is None:
        return Xtrain,mean_xtrain, std_xtrain
    else:
        temp2 = Xtest[..., quant_index]
        if type(temp2) != torch.Tensor:
            temp2 = temp2.astype(float)
        temp2 = (temp2 - mean_xtrain)/std_xtrain
        Xtest[..., quant_index] = temp2
        if type(Xtest) == np.ndarray:
            Xtest = torch.from_numpy(Xtest)
        return Xtrain, Xtest, mean_xtrain, std_xtrain
    

def compute_mean_std(tensor):
    means = torch.nanmean(tensor, dim=0)

    # Mask for non-NaN elements
    mask = ~torch.isnan(tensor)
    
    # Calculating the standard deviation
    diffs = tensor - means
    diffs[~mask] = 0  # Set differences to 0 where tensor is NaN
    sum_sq = torch.sum(diffs ** 2, dim=0)
    count = mask.sum(dim=0)

    # Avoid division by zero for columns with all NaN values
    zero_count_mask = count == 0
    count[zero_count_mask] = 1  # Temporarily set to 1 to avoid division by zero

    stds = torch.sqrt(sum_sq / count)
    stds[zero_count_mask] = 0  # Set std to 0 for all-NaN columns

    return means, stds




# Copyright © 2021 by Northwestern University.
# 
# LVGP-PyTorch is copyrighted by Northwestern University. It may be freely used 
# for educational and research purposes by  non-profit institutions and US government 
# agencies only. All other organizations may use LVGP-PyTorch for evaluation purposes 
# only, and any further uses will require prior written approval. This software may 
# not be sold or redistributed without prior written approval. Copies of the software 
# may be made by a user provided that copies are not sold or distributed, and provided 
# that copies are used under the same terms and conditions as agreed to in this 
# paragraph.
# 
# As research software, this code is provided on an "as is'' basis without warranty of 
# any kind, either expressed or implied. The downloading, or executing any part of this 
# software constitutes an implicit agreement to these terms. These terms and conditions 
# are subject to change at any time without prior notice.

import math
import torch
from gpytorch.priors import Prior
from torch.distributions import constraints,Uniform,Normal
from torch.distributions.utils import broadcast_all
from numbers import Number

class MollifiedUniformPrior(Prior):
    r"""Uniform distribution that is differentiable everywhere 

    This is an approximation to the Uniform distribution which maintains differentiability by placing a 
    Gaussian distribution over points away from the original support. The density for a single dimension is

    .. math::

        p(x) &= \frac{M}{\sqrt{2\pi\sigma^2}}\exp\left[-\frac{d(x)^2}{2\sigma^2}\right], \\
        d(x) &= \begin{cases}
            a-x & x < a \\
            0 & a\leq x < b\\
            x-b & x \geq b
        \end{cases}

    :param a: lower range (inclusive)
    :type a: float or torch.Tensor

    :param b: upper range (exclusive)
    :type b: float or torch.Tensor

    :param tail_sigma: Standard deviation of the Gaussian distributions on the tails. Lower values make the 
        approximation closer to Uniform, but may increase the optimization effort. Defaults to 0.1
    :type tail_sigma: float or torch.Tensor, optional

    .. note::
        The `rsample` method for this distribution returns uniformly distributed samples from the interval `[a,b)`, 
        and **not** from the Mollified distribution. The `log_prob` method, however, returns the correct probability.
        The purpose of the priors in this package is to generate initial starting points for optimizing
        hyperparameters and for MAP estimation. 
    """
    arg_constraints = {'a':constraints.real,'b':constraints.real,'tail_sigma':constraints.positive}
    support = constraints.real
    def __init__(self,a,b,tail_sigma=0.1):
        self.a,self.b,self.tail_sigma = broadcast_all(a,b,tail_sigma)
        
        if isinstance(a,Number) or isinstance(b,Number):
            batch_shape = torch.Size()
        else:
            batch_shape = self.a.size()

        super().__init__(batch_shape)
    
    @property
    def mean(self):
        return (self.a+self.b)/2
    
    @property
    def _half_range(self):
        return (self.b-self.a)/2

    @property
    def _log_normalization_constant(self):
        return -torch.log(1+(self.b-self.a)/(math.sqrt(2*math.pi)*self.tail_sigma))

    def log_prob(self,X):
        # expression preserving gradients under automatic differentiation
        tail_dist = ((X-self.mean).abs()-self._half_range).clamp(min=0)
        return Normal(loc=torch.zeros_like(self.a),scale=self.tail_sigma).log_prob(tail_dist)+self._log_normalization_constant
    
    def rsample(self,sample_shape=torch.Size([])):
        return Uniform(self.a,self.b).rsample(sample_shape).to(self.a)

    def expand(self,expand_shape):
        batch_shape = torch.Size(expand_shape)
        return MollifiedUniformPrior(
            self.a.expand(batch_shape),
            self.b.expand(batch_shape),
            self.tail_sigma.expand(batch_shape)
        )



import torch
import numpy as np
import random
# Set the seed for reproducibility

def set_seed(seed):
    random.seed(seed)
    torch.manual_seed(seed)
    np.random.seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed_all(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed_all(seed)
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False


def one_hot_encoding(X,qual_index):
    import numpy as np
    import torch
    import torch.nn.functional as F
    X_numerical= torch.tensor(np.delete(X.numpy(), list(qual_index.keys()), 1))
    X_CAT_one=torch.zeros(X_numerical.shape[0],sum(qual_index.values()))
    # Xtrain_CAT_one=qual_index
    Xtrain_CAT_one=[]
    k=0
    for i in qual_index.keys():
        x_one_hoted=F.one_hot(X[:,i].to(torch.int64), num_classes=qual_index[i])
        # Xtrain_CAT_one[i]=x_one_hoted
        if k==0:
            X_CAT_one_all=x_one_hoted
            k=+1
        else:
            X_one_hot_all=torch.cat((X_CAT_one_all,x_one_hoted),1)


    X_final=torch.cat((X_one_hot_all.to(torch.int64),X_numerical),1)
    return X_final